;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "org-mem" "0.34.1"
  "Fast info from a large number of Org file contents."
  '((emacs          "29.1")
    (el-job         "2.7.3")
    (llama          "1.0")
    (truename-cache "0.3.2"))
  :url "https://github.com/meedstrom/org-mem"
  :commit "04abbc49d936dc632672d95851b0b1f2af4c2e3f"
  :revdesc "0.34.1-0-g04abbc49d936"
  :keywords '("text")
  :authors '(("Martin Edström" . "meedstrom@runbox.eu"))
  :maintainers '(("Martin Edström" . "meedstrom@runbox.eu")))
